#include <hidef.h> /* for EnableInterrupts macro */
#include "derivative.h" /* include peripheral declarations */

#pragma DATA_SEG __SHORT_SEG MY_ZEROPAGE
byte    duty;           // Duty value.
byte    ad_index;       // Index for ADC channel
byte    t_adc[4];       // Temporary ADC results
byte    vadc[4];        // Final ADC results
struct {
  byte    adc_ready:1;  // ADC cycle completed.
}       flags;
byte    target;         // taret value
byte    err;            // error for target
word    diff;           // Differential from TYP_DUTY
byte    duty_nxt;       // DUTY for next frame
byte    frame_count;    // Counter of frames
byte    kp;             // P factor
#pragma DATA_SEG DEFAULT

void init(void) {
  //--------------------------------------------
  // ICS Configuration.
  //--------------------------------------------
  ICSTRM = NVICSTRM;    // Adjust internal OSC
  ICSSC  = NVFTRIM;
  ICSC2_BDIV = 0;       // BUSCLK = DCO/2
  //--------------------------------------------
  // ADC Configuration.
  //--------------------------------------------
  ADCCFG
    = 0b01000000;
  //    ||||||++-- ADICLK : BUSCLK selected
  //    ||||++---- MODE   : 8-bit conversion
  //    |||+------ ADLSMP : Short sampling
  //    |++------- ADIV   : 1/4 divided clock
  //    +--------- ADLPC  : High speed conversion
  APCTL1 = 0b00001100;  // Enable AD3 and AD2
  SPMSC1_BGBE = 1;      // Enable internal band-gap
  //--------------------------------------------
  // TPM Configuration.
  //--------------------------------------------
  TPM1C0V = 4;
  TPM1C0SC
    = 0b01010000;
  //    ||||||++-- Reserved
  //    ||||++---- ELS    : No port associated
  //    ||++------ MS     : Output compare mode
  //    |+-------- CH0IE  : Enable interrupt on CH0
  //    +--------- CH0F   : N/A
  TPM1MOD = 255;        // TOF period 25.6usec
  TPM1SC
    = 0b01001001;
  //    |||||+++-- PS     : 1/2 prescaler
  //    |||++----- CLKS   : BUSCLK selected
  //    ||+------- CPWMS  : edge aligned PWM
  //    |+-------- TOIE   : Enable interrupt on TOF
  //    +--------- TOF    : N/A
  //--------------------------------------------
  // Gport Configuration.
  //--------------------------------------------
  PTCD  = 0x00;         // Output ZERO for PTC[3:0]
  PTCDS = 0x0F;         // Strong drive for PTC[3:0]
  PTBDS = 0x3C;         // Strong drive for PTB[5:2]
  GNGC  = 0xFF;         // 8-port GANG output
  //--------------------------------------------
  // Variable initialization
  //--------------------------------------------
  duty = 20;            // Starts from minimum duty
  duty_nxt = 20;        // Starts from minimum duty
  target = 77;          // Target to 0.3 VDD
  kp = 32;               // Default P factor
}

#pragma TRAP_PROC SAVE_NO_REGS
void __interrupt VectorNumber_Vtpm1ovf isr_ovf(void) {
  TPM1C0VH = 0;
  TPM1C0VL = duty;      // Set next duty ratio.
  TPM1SC_TOF = 0;       // Clear overflow flag.
  PTCD_PTCD0 = 1;       // Assert Gport
}

#pragma TRAP_PROC SAVE_NO_REGS
void __interrupt VectorNumber_Vtpm1ch0 isr_ch0(void) {
  PTCD_PTCD0 = 0;       // Negate Gport
  TPM1C0SC_CH0F = 0;    // Clear channel flag.
  
  // Multi-channel ADC state machine
  if (ad_index < 4) {
    if (ad_index < 2) {
      if (ad_index < 1) {
        // ad_index == 0
        ADCSC1 = 27;        // Start BG conversion.
        ad_index = 1;       // Set next state
      } else {
        // ad_index == 1
        t_adc[0] = ADCRL; // Get ADC value
        ad_index = 2;     // Set next state
      }
    } else {
      if (ad_index < 3) {
        // ad_index == 2
        ADCSC1 = 2;         // Start AD2 conversion.
        ad_index = 3;       // Set next state
      } else {
        // ad_index == 3
        t_adc[1] = ADCRL; // Get ADC value
        ad_index = 4;     // Set next state
      }
    }
  } else {
    if (ad_index < 6) {
      if (ad_index < 5) {
        // ad_index == 4
        ADCSC1 = 3;         // Start AD3 conversion.
        ad_index = 5;       // Set next state
      } else {
        // ad_index == 5
        t_adc[2] = ADCRL; // Get ADC value
        ad_index = 6;     // Set next state
      }
    } else {
      if (ad_index < 7) {
        // ad_index == 6
        ad_index = 7;       // Set next state
      } else {
        // ad_index >= 7
        vadc[0] = t_adc[0]; // Copy ADC results
        vadc[1] = t_adc[1]; // Copy ADC results
        vadc[2] = t_adc[2]; // Copy ADC results
        flags.adc_ready = 1;// Set READY flag
        ad_index = 0;       // Default to ZERO
      }
    }
  }
}

#define     MAX_DUTY    (132)
#define     MIN_DUTY    (20)
#define     TYP_DUTY    (84)

void main(void) {

  init();
  
  EnableInterrupts; /* enable interrupts */
  /* include your code here */

  for(;;) {
    __RESET_WATCHDOG(); /* feeds the dog */
    // Wait for a frame
    flags.adc_ready = 0;
    while (!flags.adc_ready);
    // Update DUTY
    duty = duty_nxt;
    // Update TARGET if required.
    if (frame_count == 0) {
      target = 114 - target;
    }
    // Calculate ERROR and DUTY
    if (vadc[1] < target) {
      // More DUTY expected.
      err = target - vadc[1];
      diff = err * kp >> 4;
      if (diff > MAX_DUTY-TYP_DUTY) {
        duty_nxt = MAX_DUTY;    // maximum duty
      } else {
        duty_nxt = TYP_DUTY + diff;
      }
    } else if (vadc[1] > target) {
      // Less DUTY expected.
      err = vadc[1] - target;
      diff = err * kp >> 4;
      if (diff > TYP_DUTY-MIN_DUTY) {
        duty_nxt = MIN_DUTY;    // minimum duty
      } else {
        duty_nxt = TYP_DUTY - diff;
      }
    } else {
      // Norminal DUTY expected.
      duty_nxt = TYP_DUTY;
    }
    frame_count++;
  } /* loop forever */
  /* please make sure that you never leave main */
}
